//
//  Entity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "Entity.h"

#import "vector.h"
#import "Universe.h"
#import "TextureStore.h"
#import "ResourceManager.h"

static  Universe	*data_store_universe;

@implementation Entity

// class methods, they set the underlying data_storage universe
+ (void) setDataStore:(Universe *)univ
{
	if (univ)
		data_store_universe = univ;
	//NSLog(@"--- Universe for Data Storage set to %@", univ);
}

+ (Universe *) dataStore
{
	return data_store_universe;
}

- (id) init
{    
    self = [super init];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
	//
	zero_distance = 0.0;  //  10 km
	no_draw_distance = 100000.0;  //  10 km
	//
	energy =	0.0;
	//
	collision_radius = 0.0;
	//
	collidingEntities = [[NSMutableArray alloc] initWithCapacity:16];   // alloc automatically retains
	//
	scan_class = CLASS_NO_DRAW;
	//
	universal_id = NO_TARGET;
	universe = nil;
	//
	is_smooth_shaded = NO;
    //
    n_vertices = 0;
    n_faces = 0;
    //
    displayListName = 0;
    //
    status = STATUS_DEMO;
    //
    basefile = @"No Model";
	//
    return self;
}

- (void) dealloc
{
    if (universe)	[universe release];
    if (basefile)	[basefile release];
	if (collidingEntities)	[collidingEntities release];
	[super dealloc];
}

- (void) warnAboutHostiles
{
	// do nothing for now, this can be expanded in sub classes
	NSLog(@"***** Entity does nothing in warnAboutHostiles");
}

- (void) setUniverse:(Universe *)univ
{
    if (univ)
    {
        if (universe)	[universe release];
        universe = [univ retain];
    }
	else
	{
        if (universe)	[universe release];
        universe = nil;
    }
}

- (void) setUniversal_id:(int)uid
{
	universal_id = uid;
}

- (int) universal_id
{
	return universal_id;
}

- (BOOL) isSmoothShaded
{
	return is_smooth_shaded;
}
- (void) setSmoothShaded:(BOOL) value
{
	is_smooth_shaded = value;
}

- (void) setOwner:(Entity *) ent
{
	owner = [ent universal_id];	// assume this is retained somewhere
}

- (Entity *) owner
{
	return [universe entityForUniversalID:owner];
}

- (void) setModel:(NSString *) modelName
{    
	// clear old data
	if (basefile)	[basefile release];
    basefile = [modelName retain];
	//
	[self regenerateDisplayList];
    //
    [self loadData:basefile];
    //
    [self checkNormalsAndAdjustWinding];
    //
	if (is_smooth_shaded)
		[self calculateVertexNormals];
	// set the collision radius
	//
	collision_radius = [self findCollisionRadius];
	//NSLog(@"Entity with model '%@' collision radius set to %f",modelName, collision_radius);
	//
}
- (NSString *) getModel
{
	return basefile;
}

- (void) setPosition:(Vector) posn
{
	position.x = posn.x;
	position.y = posn.y;
	position.z = posn.z;
}

- (void) setPosition:(GLfloat) x:(GLfloat) y:(GLfloat) z
{
	position.x = x;
	position.y = y;
	position.z = z;
}

- (double) getZeroDistance
{
	return zero_distance;
}

- (NSComparisonResult) compareZeroDistance:(Entity *)otherEntity;
{
	if (zero_distance > [otherEntity getZeroDistance])
		return NSOrderedAscending;
	else
		return NSOrderedDescending;
}

- (BoundingBox) getBoundingBox
{
	return boundingBox;
}

- (void) setQRotation:(Quaternion) quat
{
	q_rotation = quat;
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
}

- (Quaternion) QRotation
{
	return q_rotation;
}

- (void) setVelocity:(Vector) vel
{
	velocity = vel;
}

- (Vector) getVelocity
{
	return velocity;
}

- (double) getVelocityAsSpeed
{
	return sqrt(velocity.x * velocity.x + velocity.y * velocity.y + velocity.z * velocity.z);
}

- (void) setStatus:(int) stat
{
	status = stat;
}

- (int) getStatus
{
	return status;
}

- (void) setScanClass:(int) s_class
{
	scan_class = s_class;
}

- (int) scanClass
{
	return scan_class;
}

- (void) setEnergy:(double) amount
{
	energy = amount;
}

- (double) getEnergy
{
	return energy;
}




- (void) applyRoll:(GLfloat) roll andClimb:(GLfloat) climb
{	
	quaternion_rotate_about_z( &q_rotation, -roll);
	quaternion_rotate_about_x( &q_rotation, -climb);
	
    quaternion_normalise(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
}

- (void) applyRoll:(GLfloat) roll climb:(GLfloat) climb andYaw:(GLfloat) yaw
{
	quaternion_rotate_about_z( &q_rotation, -roll);
	quaternion_rotate_about_x( &q_rotation, -climb);
	quaternion_rotate_about_y( &q_rotation, -yaw);
	
    quaternion_normalise(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
}

- (void) moveForward:(double) amount
{
    Vector		forward = vector_forward_from_quaternion(q_rotation);

	position.x += amount * forward.x;
	position.y += amount * forward.y;
	position.z += amount * forward.z;
}

- (GLfloat *) rotationMatrix
{
    return rotMatrix;
}
- (Vector) getPosition
{
    return position;
}


- (BOOL) canCollide
{
	return YES;
}

- (double) collisionRadius
{
	return collision_radius;
}

- (void) setCollisionRadius:(double) amount
{
	collision_radius = amount;
}

- (NSMutableArray *) collisionArray
{
	return collidingEntities;
}


- (void) drawEntity:(BOOL) immediate :(BOOL) translucent
{
    // roll out each face and vertex in turn
    //
    int fi,vi;
    GLfloat mat_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
	if (is_smooth_shaded)
		glShadeModel(GL_SMOOTH);
	else
		glShadeModel(GL_FLAT);
	//glShadeModel(GL_SMOOTH);
    //
    if (!translucent)
	{
		if (basefile)
		{
			if (immediate)
			{
				for (fi = 0; fi < n_faces; fi++)
				{
					GLfloat amb_diff[] = { faces[fi].red, faces[fi].blue, faces[fi].green, 1.0};
					
					// texture
					if ((faces[fi].texName == 0)&&(faces[fi].textureFile))
					{
						// load texture into Universe texturestore
						//NSLog(@"Off to load %@",faces[fi].textureFile);
						
						if (universe)
						{
							faces[fi].texName = [[universe textureStore] getTextureNameFor:faces[fi].textureFile];
						}
					}
					
					if (faces[fi].texName != 0)
					{
						// apply texture
						glEnable(GL_TEXTURE_2D);
						glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
						glBindTexture(GL_TEXTURE_2D, faces[fi].texName);
						glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
					}
					else
					{
						glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_diff);
					}
					
					if (!is_smooth_shaded)
						glNormal3f(faces[fi].normal.x, faces[fi].normal.y, faces[fi].normal.z);
						
					glBegin(GL_POLYGON);
					for (vi = 0; vi < faces[fi].n_verts; vi++)
					{
						int v = faces[fi].vertex[vi];
						if (is_smooth_shaded)
							glNormal3f(vertex_normal[v].x,vertex_normal[v].y,vertex_normal[v].z);
						glTexCoord2f(faces[fi].s[vi],faces[fi].t[vi]);
						glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z);
					}
					glEnd();
					
					if (faces[fi].texName != 0)
					{
						// disapply texture
						glDisable(GL_TEXTURE_2D);
					}
				}
			}
			else
			{
				if (displayListName != 0)
					glCallList(displayListName);
				else
				{
					[self initialiseTextures];
					[self generateDisplayList];
					//
					//NSLog(@"Clean Data:\n%@\n", [self toString]);
					//
				}
			}
		}
	}
	glShadeModel(GL_SMOOTH);
}

- (void) initialiseTextures
{
    // roll out each face and vertex in turn
    //
    int fi;
    //
    for (fi = 0; fi < n_faces; fi++)
    {        
        // texture
        if ((faces[fi].texName == 0)&&(faces[fi].textureFile))
        {
            // load texture into Universe texturestore
            //NSLog(@"Off to load %@",faces[fi].textureFile);
            if (universe)
            {
                faces[fi].texName = [[universe textureStore] getTextureNameFor:faces[fi].textureFile];
            }
        }
    }
}

- (void) regenerateDisplayList
{
	glDeleteLists(displayListName,1);
	displayListName = 0;
}

- (void) generateDisplayList
{
    displayListName = glGenLists(1);
    if (displayListName != 0)
    {
        glNewList(displayListName, GL_COMPILE);
        [self drawEntity:YES:NO];
        glEndList();
    }
}

- (void) update:(double) delta_t
{
	if (universe)
	{
		Vector p0 = [[universe entityZero] getPosition];
		p0.x -= position.x; p0.y -= position.y;  p0.z -= position.z;
		zero_distance = p0.x*p0.x + p0.y*p0.y + p0.z*p0.z;
	}
}


- (void) loadData:(NSString *) filename
{
    NSScanner	*scanner;
    NSString	*data;
    NSMutableArray	*lines;
    BOOL	failFlag = NO;
    NSString	*failString = @"";
    int	i, j;
    
	if (data_store_universe)
	{
		if ([[data_store_universe preloadedDataFiles] objectForKey:filename])
		{
			//NSLog(@"Reusing data from [data_store_universe preloadedDataFiles]");
			data = (NSString *)[[data_store_universe preloadedDataFiles] objectForKey:filename];
		}
		else
		{
			data = [ResourceManager stringFromFilesNamed:basefile inFolder:@"Models"];
			[[data_store_universe preloadedDataFiles] setObject:data forKey:filename];
		}
	}
	else
	{
		data = [ResourceManager stringFromFilesNamed:basefile inFolder:@"Models"];
	}
    
    // strip out comments and commas between values
    //
    lines = [NSMutableArray arrayWithArray:[data componentsSeparatedByString:@"\n"]];
    for (i = 0; i < [ lines count]; i++)
    {
        NSString *line = [lines objectAtIndex:i];
        NSArray *parts;
        //
        // comments
        //
        parts = [line componentsSeparatedByString:@"#"];
        line = [parts objectAtIndex:0];
        parts = [line componentsSeparatedByString:@"//"];
        line = [parts objectAtIndex:0];
        //
        // commas
        //
        line = [[line componentsSeparatedByString:@","] componentsJoinedByString:@" "];
        //
        [lines replaceObjectAtIndex:i withObject:line];
    }
    data = [lines componentsJoinedByString:@"\n"];
    
    //NSLog(@"More data:\n%@",data);
   
    scanner = [NSScanner scannerWithString:data];
    
    // get number of vertices
    //
    [scanner setScanLocation:0];	//reset
    //[scanner scanCharactersFromSet:[NSCharacterSet whitespaceAndNewlineCharacterSet] intoString:nil];
    if ([scanner scanString:@"NVERTS" intoString:nil])
    {
        int n_v;
        if ([scanner scanInt:&n_v])
            n_vertices = n_v;
        else
        {
            failFlag = YES;
            failString = [NSString stringWithFormat:@"%@Failed to read value of NVERTS\n",failString];
        }
    }
    else
    {
        failFlag = YES;
        failString = [NSString stringWithFormat:@"%@Failed to read NVERTS\n",failString];
    }
    
    // get number of faces
    //
    //[scanner setScanLocation:0];	//reset
    if ([scanner scanString:@"NFACES" intoString:nil])
    {
        int n_f;
        if ([scanner scanInt:&n_f])
            n_faces = n_f;
        else
        {
            failFlag = YES;
            failString = [NSString stringWithFormat:@"%@Failed to read value of NFACES\n",failString];
        }
    }
    else
    {
        failFlag = YES;
        failString = [NSString stringWithFormat:@"%@Failed to read NFACES\n",failString];
    }
    
    // get vertex data
    //
    //[scanner setScanLocation:0];	//reset
    if ([scanner scanString:@"VERTEX" intoString:nil])
    {
        for (j = 0; j < n_vertices; j++)
        {
            float x, y, z;
            if (!failFlag)
            {
                if (![scanner scanFloat:&x])
                    failFlag = YES;
                if (![scanner scanFloat:&y])
                    failFlag = YES;
                if (![scanner scanFloat:&z])
                    failFlag = YES;
                if (!failFlag)
                {
                    vertices[j].x = x;	vertices[j].y = y;	vertices[j].z = z;
                }
                else
                {
                    failString = [NSString stringWithFormat:@"%@Failed to read a value for vertex[%d] in VERTEX\n", failString, j];
                }
            }
        }
    }
    else
    {
        failFlag = YES;
        failString = [NSString stringWithFormat:@"%@Failed to find VERTEX data\n",failString];
    }
    
    // get face data
    //
    //[scanner setScanLocation:0];	//reset
    if ([scanner scanString:@"FACES" intoString:nil])
    {
        for (j = 0; j < n_faces; j++)
        {
            int r, g, b;
            float nx, ny, nz;
            int n_v;
            if (!failFlag)
            {
                // colors
                //
                if (![scanner scanInt:&r])
                    failFlag = YES;
                if (![scanner scanInt:&g])
                    failFlag = YES;
                if (![scanner scanInt:&b])
                    failFlag = YES;
                if (!failFlag)
                {
                    faces[j].red = r/255.0;    faces[j].green = g/255.0;    faces[j].blue = b/255.0;
                }
                else
                {
                    failString = [NSString stringWithFormat:@"%@Failed to read a color for face[%d] in FACES\n", failString, j];
                }
                
                // normal
                //
                if (![scanner scanFloat:&nx])
                    failFlag = YES;
                if (![scanner scanFloat:&ny])
                    failFlag = YES;
                if (![scanner scanFloat:&nz])
                    failFlag = YES;
                if (!failFlag)
                {
                    faces[j].normal.x = nx;    faces[j].normal.y = ny;    faces[j].normal.z = nz;
                }
                else
                {
                    failString = [NSString stringWithFormat:@"%@Failed to read a normal for face[%d] in FACES\n", failString, j];
                }
                
                // vertices
                //
                if ([scanner scanInt:&n_v])
                {
                    faces[j].n_verts = n_v;
                }
                else
                {
                    failFlag = YES;
                    failString = [NSString stringWithFormat:@"%@Failed to read number of vertices for face[%d] in FACES\n", failString, j];
                }
                //
                if (!failFlag)
                {
                    int vi;
                    for (i = 0; i < n_v; i++)
                    {
                        if ([scanner scanInt:&vi])
                        {
                            faces[j].vertex[i] = vi;
                        }
                        else
                        {
                            failFlag = YES;
                            failString = [NSString stringWithFormat:@"%@Failed to read vertex[%d] for face[%d] in FACES\n", failString, i, j];
                        }
                    }
                }
            }
        }
    }
    else
    {
        failFlag = YES;
        failString = [NSString stringWithFormat:@"%@Failed to find FACES data\n",failString];
    }
    
    // get textures data
    //
    if ([scanner scanString:@"TEXTURES" intoString:nil])
    {
        for (j = 0; j < n_faces; j++)
        {
            NSString	*texfile;
            float	max_x, max_y;
            float	s, t;
            if (!failFlag)
            {
                // texfile
                //
                [scanner scanCharactersFromSet:[NSCharacterSet whitespaceAndNewlineCharacterSet] intoString:nil];
                if (![scanner scanUpToCharactersFromSet:[NSCharacterSet whitespaceCharacterSet] intoString:&texfile])
                {
                    failFlag = YES;
                    failString = [NSString stringWithFormat:@"%@Failed to read texture filename for face[%d] in TEXTURES\n", failString, j];
                }
                else
                {
                    faces[j].textureFile = [texfile retain];
                }
                faces[j].texName = 0;
                
                // texture size
                //
               if (!failFlag)
                {
                    if (![scanner scanFloat:&max_x])
                        failFlag = YES;
                    if (![scanner scanFloat:&max_y])
                        failFlag = YES;
                    if (failFlag)
                        failString = [NSString stringWithFormat:@"%@Failed to read texture size for max_x and max_y in face[%d] in TEXTURES\n", failString, j];
                }
                
                // vertices
                //
                if (!failFlag)
                {
                    for (i = 0; i < faces[j].n_verts; i++)
                    {
                        if (![scanner scanFloat:&s])
                            failFlag = YES;
                        if (![scanner scanFloat:&t])
                            failFlag = YES;
                        if (!failFlag)
                        {
                            faces[j].s[i] = s / max_x;    faces[j].t[i] = t / max_y;
                            
                            //NSLog(@" st %f %f", faces[j].s[i], faces[j].t[i]);
                        }
                        else
                            failString = [NSString stringWithFormat:@"%@Failed to read s t coordinates for vertex[%d] in face[%d] in TEXTURES\n", failString, i, j];
                    }
                }
            }
        }
    }
    else
    {
        failFlag = YES;
        failString = [NSString stringWithFormat:@"%@Failed to find TEXTURES data\n",failString];
    }
	
	// check normals before creating new textures
	//
    [self checkNormalsAndAdjustWinding];
    
    if ((failFlag)&&([failString rangeOfString:@"TEXTURES"].location != NSNotFound))
    {
        //NSLog(@"Off to make new textures!");
        [self fakeTexturesWithImageFile:@"metal.png" andMaxSize:NSMakeSize(256.0,256.0)];
		
		// dump out data for ships with faked textures
		//if ([self isKindOfClass:[ShipEntity class]])
		//	//NSLog(@"Faked Texture coordinates for this model :\n\n%@\n\n", [self toString]);
    }
    
    if (failFlag)
        NSLog(failString);
	
	// set the collision radius
	//
	collision_radius = [self findCollisionRadius];
	//
}

- (void) checkNormalsAndAdjustWinding
{
    Vector calculatedNormal;
    int i, j;
    for (i = 0; i < n_faces; i++)
    {
        Vector v0, v1, v2, norm;
        v0 = vertices[faces[i].vertex[0]];
        v1 = vertices[faces[i].vertex[1]];
        v2 = vertices[faces[i].vertex[2]];
        norm = faces[i].normal;
		calculatedNormal = normal_to_surface (v2, v1, v0);
        if ((norm.x == 0.0)&&(norm.y == 0.0)&&(norm.z == 0.0))
		{
			//NSLog(@"Using calculated normal for face %d", i);
			faces[i].normal = normal_to_surface (v0, v1, v2);
			norm = normal_to_surface (v0, v1, v2);
		}
        if ((norm.x*calculatedNormal.x < 0)||(norm.y*calculatedNormal.y < 0)||(norm.z*calculatedNormal.z < 0))
        {
            // normal lies in the WRONG direction!
            // reverse the winding
            int v[faces[i].n_verts];
            GLfloat s[faces[i].n_verts];
            GLfloat t[faces[i].n_verts];
            
            //
            //NSLog(@"Normal pointing the wrong way for winding on face %d", i);
            //
            
            for (j = 0; j < faces[i].n_verts; j++)
            {
            	v[j] = faces[i].vertex[faces[i].n_verts - 1 - j];
            	s[j] = faces[i].s[faces[i].n_verts - 1 - j];
            	t[j] = faces[i].t[faces[i].n_verts - 1 - j];
            }
            for (j = 0; j < faces[i].n_verts; j++)
            {
            	faces[i].vertex[j] = v[j];
                faces[i].s[j] = s[j];
                faces[i].t[j] = t[j];
            }
        }
    }
}

- (void) calculateVertexNormals
{
	int i,j,k;
	for (i = 0; i < n_vertices; i++)
	{
		int shared_faces = 0;
		Vector normal_sum;
		normal_sum.x = 0.0;	normal_sum.y = 0.0;	normal_sum.z = 0.0;
		for (j = 0; j < n_faces; j++)
		{
			BOOL is_shared = NO;
			for (k = 0; (k < faces[j].n_verts)&&(!is_shared); k++)
				is_shared = (faces[j].vertex[k] == i);
			if (is_shared)
			{
				normal_sum.x += faces[j].normal.x;	normal_sum.y += faces[j].normal.y;	normal_sum.z += faces[j].normal.z;
				shared_faces++;
			}
		}
		normal_sum = unit_vector(&normal_sum);
		vertex_normal[i].x = normal_sum.x;
		vertex_normal[i].y = normal_sum.y;
		vertex_normal[i].z = normal_sum.z;
	}
}



- (double) findCollisionRadius
{
    int i;
	double d_squared,result;
	
	result = 0.0;
	bounding_box_reset(&boundingBox);
	
    for (i = 0; i < n_vertices; i++)
    {
        d_squared = vertices[i].x*vertices[i].x + vertices[i].y*vertices[i].y + vertices[i].z*vertices[i].z;
        if (d_squared > result)
			result = d_squared;
		bounding_box_add_vector(&boundingBox,vertices[i]);
    }
	//NSLog(@"Bounding Box ((%3.2f,%3.2f),(%3.2f,%3.2f),(%3.2f,%3.2f))\nCollision Radius %3.2f",
	//	boundingBox.min_x, boundingBox.max_x, boundingBox.min_y, boundingBox.max_y,
	//	boundingBox.min_z, boundingBox.max_z, result);
	no_draw_distance = result * NO_DRAW_DISTANCE_FACTOR * NO_DRAW_DISTANCE_FACTOR;
	return sqrt(result);
}


- (BoundingBox) findBoundingBoxRelativeTo:(Entity *)other InVectors:(Vector) _i :(Vector) _j :(Vector) _k
{
	Vector pv, rv;
	Vector  rpos = position;
	Vector  opv = [other getPosition];
	rpos.x -= opv.x;	rpos.y -= opv.y;	rpos.z -= opv.z;
	rv.x = dot_product(_i,rpos);
	rv.y = dot_product(_j,rpos);
	rv.z = dot_product(_k,rpos);
	BoundingBox result;
	bounding_box_reset_to_vector(&result,rv);
	int i;
    for (i = 0; i < n_vertices; i++)
    {
		pv.x = rpos.x + vertices[i].x;
		pv.y = rpos.y + vertices[i].y;
		pv.z = rpos.z + vertices[i].z;
		rv.x = dot_product(_i,pv);
		rv.y = dot_product(_j,pv);
		rv.z = dot_product(_k,pv);
		bounding_box_add_vector(&result,rv);
    }
	
	return result;
}

- (BOOL) checkCloseCollisionWith:(Entity *)other
{
	return YES;
}



- (void) takeEnergyDamage:(double) amount from:(Entity *) ent becauseOf:(Entity *) other
{
}

- (NSString *) toString
{
    // produce a file from the original data
    int i,j, r,g,b;
    NSString *result;
    NSString *boilerplate = @"# This is a file adapted from the model files for Java Elite\n# which in turn are based on the data released by Ian Bell\n# in the file b7051600.zip at\n# http://www.users.waitrose.com/~elitearc2/elite/archive/b7051600.zip\n#";
    result = [NSString stringWithFormat:@"%@\n# %@\n#\n\nNVERTS %d\nNFACES %d\n\nVERTEX\n", boilerplate, basefile, n_vertices, n_faces];
    for (i = 0; i < n_vertices; i++)
    {
        result = [NSString stringWithFormat:@"%@%f,\t%f,\t%f\n", result, vertices[i].x, vertices[i].y, vertices[i].z];
        if ((i % 5)==4)
            result = [NSString stringWithFormat:@"%@\n", result];
    }
    result = [NSString stringWithFormat:@"%@\nFACES\n", result];
    //
    //NSLog(result);
    //
    for (j = 0; j < n_faces; j++)
    {
        r = (int)(faces[j].red * 255.0);	g = (int)(faces[j].green * 255.0);	b = (int)(faces[j].blue * 255.0);
        result = [NSString stringWithFormat:@"%@%d, %d, %d,\t", result, r, g, b];
        result = [NSString stringWithFormat:@"%@%f, %f, %f,\t", result, faces[j].normal.x, faces[j].normal.y, faces[j].normal.z];
        result = [NSString stringWithFormat:@"%@%d,\t", result, faces[j].n_verts];
        for (i = 0; i < faces[j].n_verts; i++)
        {
            result = [NSString stringWithFormat:@"%@%d ", result, faces[j].vertex[i]];
        }
        result = [NSString stringWithFormat:@"%@\n", result];
    }
    if (universe)
    {
        result = [NSString stringWithFormat:@"%@\nTEXTURES\n", result];
        for (j = 0; j < n_faces; j++)
        {
            NSSize	texSize = [[universe textureStore] getSizeOfTexture:faces[j].textureFile];
            result = [NSString stringWithFormat:@"%@%@\t%d %d", result, faces[j].textureFile, (int)texSize.width, (int)texSize.height];
            for (i = 0; i < faces[j].n_verts; i++)
            {
                int s = (int)(faces[j].s[i] * texSize.width);
                int t = (int)(faces[j].t[i] * texSize.height);
                result = [NSString stringWithFormat:@"%@\t%d %d", result, s, t];
            }
            result = [NSString stringWithFormat:@"%@\n", result];
        }
    }
    result = [NSString stringWithFormat:@"%@\nEND\n", result];
    
    return result;
}

- (void) fakeTexturesWithImageFile: (NSString *) textureFile andMaxSize:(NSSize) maxSize
{
    int i, j, k;
    Vector vec;
    int nf = 0;
    int		fi[MAX_FACES_PER_ENTITY];
    float	max_s, min_s, max_t, min_t, st_width, st_height;
    float	tolerance;
    Face	fa[MAX_FACES_PER_ENTITY];
    int		faces_to_match;
    BOOL	face_matched[MAX_FACES_PER_ENTITY];
    
    tolerance = 1.00;
    faces_to_match = n_faces;
    for (i = 0; i < n_faces; i++)
    {
	    face_matched[i] = NO;
    }
	while (faces_to_match > 0)
    {
        tolerance -= 0.05;
        
        // Top (+y) first
        vec.x = 0.0;	vec.y = 1.0;	vec.z = 0.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = vertices[faces[i].vertex[j]].x;
                    t = vertices[faces[i].vertex[j]].z;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        //
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        //
        //NSLog(@"TOP st_width %f st_height %f maxSize.height %f maxSize.width %f", st_width, st_height, maxSize.width, maxSize.height);
        //
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"top_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = vertices[faces[i].vertex[k]].x;
                t = vertices[faces[i].vertex[k]].z;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
				//
				// TESTING
				//
				fa[i].t[k] = maxSize.height - fa[i].t[k];	// REVERSE t locations
                //
                //NSLog(@"%f, %f", fa[i].s[k], fa[i].t[k]);
                //
            }
        }
        
        // Bottom (-y)
        vec.x = 0.0;	vec.y = -1.0;	vec.z = 0.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = -vertices[faces[i].vertex[j]].x;
                    t = -vertices[faces[i].vertex[j]].z;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"bottom_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = -vertices[faces[i].vertex[k]].x;
                t = -vertices[faces[i].vertex[k]].z;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
            }
        }
        
        // Right (+x)
        vec.x = 1.0;	vec.y = 0.0;	vec.z = 0.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = vertices[faces[i].vertex[j]].z;
                    t = vertices[faces[i].vertex[j]].y;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"right_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = vertices[faces[i].vertex[k]].z;
                t = vertices[faces[i].vertex[k]].y;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
            }
        }
        
        // Left (-x)
        vec.x = -1.0;	vec.y = 0.0;	vec.z = 0.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = -vertices[faces[i].vertex[j]].z;
                    t = -vertices[faces[i].vertex[j]].y;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"left_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = -vertices[faces[i].vertex[k]].z;
                t = -vertices[faces[i].vertex[k]].y;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
            }
        }
        
        // Front (+z)
        vec.x = 0.0;	vec.y = 0.0;	vec.z = 1.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = vertices[faces[i].vertex[j]].x;
                    t = vertices[faces[i].vertex[j]].y;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"front_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = vertices[faces[i].vertex[k]].x;
                t = vertices[faces[i].vertex[k]].y;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
            }
        }
        
        // Back (-z)
        vec.x = 0.0;	vec.y = 0.0;	vec.z = -1.0;
        // build list of faces that face in that direction...
        nf = 0;
        max_s = -999999.0; min_s = 999999.0;
        max_t = -999999.0; min_t = 999999.0;
        for (i = 0; i < n_faces; i++)
        {
            float s, t;
            float g = dot_product(vec, faces[i].normal) * sqrt(2.0);
            if ((g >= tolerance)&&(!face_matched[i]))
            {
                fi[nf++] = i;
                face_matched[i] = YES;
                faces_to_match--;
                for (j = 0; j < faces[i].n_verts; j++)
                {
                    s = -vertices[faces[i].vertex[j]].x;
                    t = -vertices[faces[i].vertex[j]].y;
                    max_s = (max_s > s) ? max_s:s ;	min_s = (min_s < s) ? min_s:s ;
                    max_t = (max_t > t) ? max_t:t ;	min_t = (min_t < t) ? min_t:t ;
                }
            }
        }
        st_width = max_s - min_s;
        st_height = max_t - min_t;
        for (j = 0; j < nf; j++)
        {
            i = fi[j];
            //fa[i] = faces[i];
            fa[i].textureFile = [NSString stringWithFormat:@"back_%@", textureFile];
            for (k = 0; k < faces[i].n_verts; k++)
            {
                float s, t;
                s = -vertices[faces[i].vertex[k]].x;
                t = -vertices[faces[i].vertex[k]].y;
                fa[i].s[k] = (s - min_s) * maxSize.width / st_width;
                fa[i].t[k] = (t - min_t) * maxSize.height / st_height;
            }
        }
        //NSLog(@"%d / %d faces matched at tolerance: %f", n_faces - faces_to_match, n_faces, tolerance);
    }
    
    for (i = 0; i < n_faces; i++)
    {
        NSString *result;
        faces[i].textureFile = [fa[i].textureFile retain];
		faces[i].texName = 0;
        for (j = 0; j < faces[i].n_verts; j++)
        {
            //
            //NSLog(@"face[%d] %f, %f", i, fa[i].s[j], fa[i].t[j]);
            //
            faces[i].s[j] = fa[i].s[j] / maxSize.width;
            faces[i].t[j] = fa[i].t[j] / maxSize.height;
        }
        result = [NSString stringWithFormat:@"%@\t%d %d", faces[i].textureFile, (int)maxSize.width, (int)maxSize.height];
        //NSLog(@"face[%d] : %@", i, result);
    }
    
}

@end
